# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib.api.definitions import portbindings
from neutron_lib import constants

from neutron.plugins.ml2.drivers import mech_agent


AGENT_TYPE_ZVM = 'z/VM agent'
VIF_TYPE_ZVM = 'zvm'


class ZvmMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using zvm vswitch agent.

    The ZvmMechanismDriver integrates the ml2 plugin with the
    z/VM L2 agent. Port binding with this driver requires the
    z/VM vswitch agent to be running on the port's host, and that agent
    to have connectivity to at least one segment of the port's network.
    """

    def __init__(self):
        super(ZvmMechanismDriver, self).__init__(
            AGENT_TYPE_ZVM,
            VIF_TYPE_ZVM,
            {portbindings.CAP_PORT_FILTER: False})

    def get_allowed_network_types(self, agent=None):
        return [constants.TYPE_LOCAL, constants.TYPE_FLAT,
                constants.TYPE_VLAN]

    def get_mappings(self, agent):
        return agent['configurations'].get('vswitch_mappings', {})



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for neutron z/VM driver
"""

import mock
from oslo_config import cfg

from neutron.plugins.zvm.agent import zvm_neutron_agent
from neutron.tests import base

SDK_URL = 'https://10.10.10.1:8080'
FLAT_NETWORKS = ['flat_net1']
VLAN_NETWORKS = ['vlan_net1:100:500']
NET_UUID = 'zvm-net-uuid'
PORT_UUID = 'zvm-port-uuid'


class FakeLoopingCall(object):
    def __init__(self, fake_time):
        self.fake_time = fake_time

    def start(self, interval=0):
        self.fake_time()


class TestZVMNeutronAgent(base.BaseTestCase):

    def setUp(self):
        super(TestZVMNeutronAgent, self).setUp()
        self.addCleanup(cfg.CONF.reset)
        cfg.CONF.set_override('cloud_connector_url', SDK_URL,
                              group='AGENT')
        cfg.CONF.set_override('rpc_backend',
                              'neutron.openstack.common.rpc.impl_fake')
        cfg.CONF.set_override('flat_networks', FLAT_NETWORKS,
                              group='ml2_type_flat')
        cfg.CONF.set_override('network_vlan_ranges', VLAN_NETWORKS,
                              group='ml2_type_vlan')

        mock.patch('neutron.openstack.common.loopingcall.'
                   'FixedIntervalLoopingCall',
                  new=FakeLoopingCall)

        with mock.patch(
                'neutron.plugins.zvm.common.utils.'
                'zVMConnectorRequestHandler') as mock_SDKReq:
            with mock.patch(
                 'neutron.plugins.zvm.common.utils.zvmUtils') as mock_Utils:
                instance = mock_SDKReq.return_value
                instance.call = mock.MagicMock(return_value={
                                                    'ipl_time': 'fake_time',
                                                    'zvm_host': 'TEST'})
                net_attrs = {'fake_uuid1': {
                       'vswitch': 'fake_vsw', 'userid': 'fake_user1'}}

                utils_ins = mock_Utils.return_value
                utils_ins.get_port_map = mock.MagicMock(
                                                return_value=net_attrs)

                self.agent = zvm_neutron_agent.zvmNeutronAgent()
                self.agent.plugin_rpc = mock.Mock()
                self.agent.context = mock.Mock()
                self.agent.agent_id = mock.Mock()

    def test_port_bound_vlan(self):
        vid = 100
        with mock.patch.object(zvm_neutron_agent, "LOG") as log:
            self._test_port_bound('vlan', vid)
            log.info.assert_called_with('Binding VLAN, VLAN ID: '
                                        '%(segmentation_id)s, port_id: '
                                        '%(port_id)s',
                                        {'segmentation_id': 100,
                                         'port_id': 1000})

    def test_port_bound_flat(self):
        with mock.patch.object(zvm_neutron_agent, "LOG") as log:
            self._test_port_bound('flat')
            log.info.assert_called_with('Bind %s port done', 1000)

    def _test_port_bound(self, network_type, vid=None):
        port = 1000
        net_uuid = NET_UUID
        enable_vlan = False

        if network_type == 'vlan':
            enable_vlan = True

        with mock.patch.object(self.agent._requesthandler, "call") as call:
            self.agent.port_bound(port, net_uuid, network_type, None,
                                  vid, 'fake_user')
            call.assert_any_call("vswitch_grant_user", None, 'fake_user')
            if enable_vlan:
                call.assert_any_call("vswitch_set_vlan_id_for_user",
                                     None, 'fake_user', int(vid))

    def test_port_unbound(self):
        # port_unbound just call utils.revoke_user, revoke_user is covered
        # in test_zvm_utils
        pass

    def test_treat_devices_added_returns_true_for_missing_device(self):
        attrs = {'get_device_details.side_effect': Exception()}
        self.agent.plugin_rpc.configure_mock(**attrs)
        # no exception should be raised
        self.agent._treat_devices_added([])

    def test_treat_devices_added_down_port(self):
        details = dict(port_id='added_port_down', physical_network='vsw',
                       segmentation_id='10', network_id='fake_net',
                       mac_address='00:11:22:33:44:55',
                       network_type='flat', admin_state_up=False)
        attrs = {'get_device_details.return_value': details}
        self.agent.plugin_rpc.configure_mock(**attrs)
        with mock.patch.object(self.agent, "_treat_vif_port",
                    mock.Mock(return_value=('fake_user'))):
            self.agent._treat_devices_added(['added_port_down'])
            self.assertTrue(self.agent.plugin_rpc.update_device_down.called)

    def test_treat_devices_added_up_port(self):
        details = dict(port_id='added_port', physical_network='vsw',
                       segmentation_id='10', network_id='fake_net',
                       mac_address='00:11:22:33:44:55',
                       network_type='flat', admin_state_up=True)
        call_ret = mock.MagicMock(return_value=[
                        {'userid': 'fake_user', 'interface': 'nic',
                         'switch': 'vs', 'port': 'nic_id', 'comments': None}])
        attrs = {'get_device_details.return_value': details}
        self.agent.plugin_rpc.configure_mock(**attrs)
        with mock.patch.object(self.agent, "_treat_vif_port",
                    mock.Mock(return_value=('fake_user'))):
            with mock.patch.object(self.agent._requesthandler, "call",
                                   call_ret):
                self.agent._treat_devices_added(['added_port'])
                self.assertTrue(self.agent.plugin_rpc.
                                get_device_details.called)

    def test_treat_devices_added_missing_port_id(self):
        details = mock.MagicMock()
        details.__contains__.side_effect = lambda x: False
        attrs = {'get_device_details.return_value': details}
        self.agent.plugin_rpc.configure_mock(**attrs)
        with mock.patch.object(zvm_neutron_agent, "LOG") as log:
            self.agent._treat_devices_added(['unknown_port'])
            self.assertTrue(log.warning.called)

    def test_treat_devices_removed_returns_true_for_missing_device(self):
        attrs = {'update_device_down.side_effect': Exception()}
        self.agent.plugin_rpc.configure_mock(**attrs)
        devices = ['fake_uuid1']
        with mock.patch.object(zvm_neutron_agent, "LOG") as log:
            self.agent._treat_devices_removed(devices)
            self.assertTrue(log.exception.called)

    def test_treat_devices_removed(self):
        devices = ['unknown_port', 'fake_uuid1']
        with mock.patch.object(zvm_neutron_agent, "LOG") as log:
            self.agent._treat_devices_removed(devices)
            log.warning.assert_called_with('Can\'t find port %s in zvm agent',
                                        'unknown_port')
            self.assertTrue(self.agent.plugin_rpc.update_device_down.called)

    # Test agent state report
    def test_report_state(self):
        with mock.patch.object(self.agent.state_rpc,
                               "report_state") as report_st:
            self.agent._report_state()
            report_st.assert_called_with(self.agent.context,
                                         self.agent.agent_state)
            self.assertNotIn("start_flag", self.agent.agent_state)

    def test_port_update_up(self):
        sdk_req_resp = []
        sdk_req_resp.append([{'userid': 'fake_user1', 'interface': '1000',
                              'switch': None, 'port': 'fake_uuid1',
                              'comments': None}])
        sdk_req_resp.append('')
        call_ret = mock.MagicMock(side_effect=sdk_req_resp)

        with mock.patch.object(self.agent.plugin_rpc,
                        "update_device_up") as rpc:
            with mock.patch.object(self.agent._requesthandler, "call",
                                   call_ret):
                self.agent.port_update(None, port={'id': 'fake_uuid1',
                                                   'admin_state_up': True})
                self.assertTrue(rpc.called)

    def test_port_update_down(self):
        sdk_req_resp = []
        sdk_req_resp.append([{'userid': 'fake_user1', 'interface': '1000',
                              'switch': None, 'port': 'fake_uuid1',
                              'comments': None}])
        sdk_req_resp.append('')
        call_ret = mock.MagicMock(side_effect=sdk_req_resp)

        with mock.patch.object(self.agent.plugin_rpc,
                        "update_device_down") as rpc:
            with mock.patch.object(self.agent._requesthandler, "call",
                                   call_ret):
                self.agent.port_update(None, port={'id': 'fake_uuid1',
                                                   'admin_state_up': False})
                self.assertTrue(rpc.called)

    def test_treat_vif_port_admin_true(self):
        call_ret = mock.MagicMock(return_value=[
            {'userid': 'fake_user', 'interface': 'nic',
             'switch': 'vs', 'port': 'nic_id', 'comments': None}])

        with mock.patch.object(self.agent, "port_bound") as bound:
            with mock.patch.object(self.agent._requesthandler, "call",
                                   call_ret):
                self.agent._treat_vif_port('port_id', 'network_id', 'flat',
                                           'vsw1', '10', True)
                self.assertTrue(bound.called)

    def test_treat_vif_port_admin_false(self):
        sdk_req_resp = []
        sdk_req_resp.append([{'userid': 'fake_user', 'interface': 'nic',
                              'switch': 'vs', 'port': 'nic_id',
                              'comments': None}])
        sdk_req_resp.append('')
        call_ret = mock.MagicMock(side_effect=sdk_req_resp)

        with mock.patch.object(self.agent._requesthandler, "call", call_ret):
            self.agent._treat_vif_port('port_id', 'network_id', 'flat',
                                       'vsw1', '10', False)
            call_ret.assert_called_with("vswitch_grant_user", "vsw1",
                                        'fake_user')

    def test_handle_restart_zvm(self):
        host_info = mock.MagicMock(return_value={
                                            'ipl_time': "zvm uptime 2"})
        port_map = mock.MagicMock()

        with mock.patch.object(self.agent._utils, "get_port_map", port_map):
            with mock.patch.object(self.agent._requesthandler, "call",
                                   host_info):
                self.agent._restart_handler.send(None)
                host_info.assert_called_with('host_get_info')
                port_map.assert_called_with()



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron._i18n import _
from neutron.conf.agent import common as config


agent_opts = [
    cfg.IntOpt(
        'polling_interval',
        default=2,
        help=_("The number of seconds the agent will wait between "
        "polling for local device changes.")),
    cfg.URIOpt('cloud_connector_url',
               schemes=['http', 'https'],
               sample_default='http://zvm.example.org:8080/',
               help="""
URL to be used to communicate with z/VM Cloud Connector.
"""),
    cfg.StrOpt('zvm_cloud_connector_ca_file',
               default=None,
               help="""
CA certificate file to be verified in httpd server.
A string, it must be a path to a CA bundle to use.
"""),
    cfg.StrOpt('zvm_cloud_connector_token_file',
               default=None,
               help="""
Token file that contains admin-token to access sdk http server.
"""),
]

CONF = cfg.CONF
CONF.register_opts(agent_opts, "AGENT")
config.register_agent_state_opts_helper(cfg.CONF)
config.register_root_helper(cfg.CONF)



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import sys
import time

from neutron_lib.agent import topics
from neutron_lib import constants as q_const
from neutron_lib import context
from oslo_log import log as logging
from oslo_service import loopingcall

from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config
from neutron.plugins.ml2.drivers.zvm import mech_zvm
from neutron.plugins.zvm.agent import zvm_network
from neutron.plugins.zvm.common import config as cfg
from neutron.plugins.zvm.common import exception
from neutron.plugins.zvm.common import utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def restart_wrapper(func):
    def wrapper(*args, **kw):
        gen = func(*args, **kw)
        gen.next()
        return gen
    return wrapper


class zvmNeutronAgent(object):
    RPC_API_VERSION = '1.1'

    def __init__(self):
        super(zvmNeutronAgent, self).__init__()
        self._requesthandler = utils.zVMConnectorRequestHandler()
        self._utils = utils.zvmUtils()
        self._polling_interval = CONF.AGENT.polling_interval
        self._host = self._requesthandler.call('host_get_info').get(
                                        'zvm_host') or CONF.host
        self._port_map = {}

        zvm_net = zvm_network.zvmNetwork()
        self.agent_state = {
            'binary': 'neutron-zvm-agent',
            'host': self._host,
            'topic': q_const.L2_AGENT_TOPIC,
            'configurations': {'vswitch_mappings': zvm_net.get_network_maps()},
            'agent_type': mech_zvm.AGENT_TYPE_ZVM,
            'start_flag': True}
        self._setup_server_rpc()
        self._restart_handler = self._handle_restart()

    def _setup_server_rpc(self):
        self.agent_id = 'zvm_agent_%s' % self._host
        self.topic = topics.AGENT
        self.plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)

        self.context = context.get_admin_context_without_session()

        self.endpoints = [self]
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.NETWORK, topics.DELETE]]
        self.connection = agent_rpc.create_consumers(self.endpoints,
                                                     self.topic,
                                                     consumers)

        report_interval = CONF.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

    def _report_state(self):
        try:
            self.state_rpc.report_state(self.context, self.agent_state)
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception("Failed reporting state!")

    def network_delete(self, context, network_id=None):
        LOG.info("Network delete received. UUID: %s", network_id)

    def port_update(self, context, **kwargs):
        port = kwargs.get('port')
        LOG.debug("Port update received. UUID: %s", port)

        if not self._port_map or not port['id'] in self._port_map.keys():
            # update a port which is not coupled to any NIC, nothing
            # to do for a user based vswitch
            return

        vswitch = self._port_map[port['id']]['vswitch']
        userid = self._port_map[port['id']]['userid']
        if port['admin_state_up']:
            nic_info = self._requesthandler.call('guests_get_nic_info',
                                                 nic_id=port['id'])
            if ((len(nic_info) != 1) or
                (len(nic_info[0]) != 5) or
                (not nic_info[0]['interface'])):
                raise exception.zVMInvalidDataError(msg=('Cannot get vdev '
                                'for user %s, couple to port %s, '
                                'SDK output is %s') %
                                (userid, port['id'], nic_info))
            else:
                vdev = nic_info[0]['interface']
            self._requesthandler.call('guest_nic_couple_to_vswitch', userid,
                                      vdev, vswitch)
            self.plugin_rpc.update_device_up(self.context, port['id'],
                                             self.agent_id)
        else:
            nic_info = self._requesthandler.call('guests_get_nic_info',
                                                 nic_id=port['id'])
            if ((len(nic_info) != 1) or
                (len(nic_info[0]) != 5) or
                (not nic_info[0]['interface'])):
                raise exception.zVMInvalidDataError(msg=('Cannot get vdev '
                                'for user %s, uncouple port %s, '
                                'SDK output is %s') %
                                (userid, port['id'], nic_info))
            else:
                vdev = nic_info[0]['interface']
            self._requesthandler.call('guest_nic_uncouple_from_vswitch',
                                      userid, vdev)
            self.plugin_rpc.update_device_down(self.context, port['id'],
                                               self.agent_id)

    def port_bound(self, port_id, net_uuid,
                   network_type, physical_network, segmentation_id, userid):
        LOG.info("Start to bind port port_id:%(port_id)s, "
                 "net_uuid:%(net_uuid)s, network_type: %(network_type)s, "
                 "physical_network: %(physical_network)s, "
                 "userid: %(userid)s, segmentation_id:%(seg_id)s",
                 {'port_id': port_id, 'net_uuid': net_uuid,
                  'network_type': network_type,
                  'physical_network': physical_network,
                  'seg_id': segmentation_id,
                  'userid': userid})

        self._requesthandler.call('vswitch_grant_user',
                                  physical_network, userid)
        if network_type == q_const.TYPE_VLAN:
            LOG.info('Binding VLAN, VLAN ID: %(segmentation_id)s, '
                     'port_id: %(port_id)s',
                     {'segmentation_id': segmentation_id,
                      'port_id': port_id})
            self._requesthandler.call('vswitch_set_vlan_id_for_user',
                                      physical_network, userid,
                                      int(segmentation_id))
        else:
            LOG.info('Bind %s port done', port_id)

    def port_unbound(self, port_id):
        LOG.info("Unbinding port %s", port_id)
        # uncouple is not necessary, because revoke user will uncouple it
        # automatically.
        self._requesthandler.call('vswitch_revoke_user',
                                  self._port_map[port_id]['vswitch'],
                                  self._port_map[port_id]['userid'])

    def _update_ports(self, registered_ports):
        nic_info = self._requesthandler.call('guests_get_nic_info')
        ports = set()
        for p in nic_info:
            if p['port'] is not None:
                new_port_id = p['port']
                ports.add(new_port_id)
        if ports == registered_ports:
            return

        added = ports - registered_ports
        removed = registered_ports - ports
        return {'current': ports, 'added': added, 'removed': removed}

    def _treat_vif_port(self, port_id, network_id, network_type,
                        physical_network, segmentation_id,
                        admin_state_up):
        nic_info = self._requesthandler.call('guests_get_nic_info',
                                             nic_id=port_id)
        if ((len(nic_info) != 1) or
            (len(nic_info[0]) != 5) or
            (not nic_info[0]['userid'])):
            raise exception.zVMInvalidDataError(msg=('Cannot get userid '
                                'for port %s, SDK output is %s') %
                                (port_id, nic_info))
        else:
            userid = nic_info[0]['userid']

        LOG.info("Update port for user:%s" % userid)
        if admin_state_up:
            self.port_bound(port_id, network_id, network_type,
                            physical_network, segmentation_id,
                            userid)
        else:
            self._requesthandler.call('vswitch_grant_user',
                                      physical_network, userid)
        return userid

    def _treat_devices_added(self, devices):
        nics_info = {}
        for device in devices:
            LOG.info("Adding port %s" % device)
            try:
                details = self.plugin_rpc.get_device_details(self.context,
                                                             device,
                                                             self.agent_id,
                                                             self._host)
            except Exception:
                LOG.info("Unable to get port details for %s:", device)
                continue

            try:
                if 'port_id' in details:
                    LOG.info("Port %(device)s updated. "
                             "Details: %(details)s",
                             {'device': device, 'details': details})
                    userid = self._treat_vif_port(
                                     details['port_id'],
                                     details['network_id'],
                                     details['network_type'],
                                     details['physical_network'],
                                     details['segmentation_id'],
                                     details['admin_state_up'])
                    # add device done, keep port map info
                    self._port_map[device] = {}
                    self._port_map[device]['userid'] = userid
                    self._port_map[device]['nodename'] = userid
                    self._port_map[device]['vswitch'] = details[
                                                        'physical_network']
                    self._port_map[device]['vlan_id'] = details[
                                                        'segmentation_id']

                    if details.get('admin_state_up'):
                        mac = ''.join(details['mac_address'].split(':'))[6:]
                        if not nics_info.get(userid):
                            nics_info[userid] = []
                        nics_info[userid].append(
                                    {'port_id': details['port_id'],
                                    'vswitch': details['physical_network'],
                                    'mac': mac})

                        LOG.debug("Adding NICs for %(userid)s, info: %(nic)s",
                                  {'userid': userid, 'nic': nics_info[userid]})
                        nic_info = self._requesthandler.call(
                                                    'guests_get_nic_info',
                                                    nic_id=details['port_id'])
                        if ((len(nic_info) != 1) or
                            (len(nic_info[0]) != 5) or
                            (not nic_info[0]['interface'])):
                            raise exception.zVMInvalidDataError(msg=('Cannot '
                                'get vdev for port %s, SDK output is %s') %
                                (details['port_id'], nic_info))
                        else:
                            vdev = nic_info[0]['interface']
                        self._requesthandler.call(
                                            'guest_nic_couple_to_vswitch',
                                            userid, vdev,
                                            details['physical_network'])

                        LOG.debug("New added NIC info: %s", nics_info[userid])

                        LOG.info("Setting status for %s to UP", device)
                        self.plugin_rpc.update_device_up(
                            self.context, device, self.agent_id, self._host)
                    else:
                        LOG.info("Setting status for %s to DOWN", device)
                        self.plugin_rpc.update_device_down(
                            self.context, device, self.agent_id, self._host)

                else:
                    LOG.warning("Device %(device)s not defined on "
                                "Neutron server, The output detail is "
                                "%(details)s",
                                {'device': device, 'details': details})
                    continue
            except Exception as e:
                LOG.exception("Can not add device %(device)s: %(msg)s",
                              {'device': device, 'msg': e})
                continue

    def _treat_devices_removed(self, devices):
        for device in devices:
            LOG.info("Removing port %s", device)
            try:
                if not self._port_map or device not in self._port_map:
                    LOG.warning("Can't find port %s in zvm agent", device)
                    continue

                if self._port_map[device]['vswitch']:
                    self.port_unbound(device)
                self.plugin_rpc.update_device_down(self.context,
                                                   device,
                                                   self.agent_id)
                del self._port_map[device]
            except Exception as e:
                LOG.exception("Removing port failed %(device)s: %(msg)s",
                              {'device': device, 'msg': e})
                continue

    def _process_network_ports(self, port_info):
        if len(port_info['added']):
            self._treat_devices_added(port_info['added'])
        if len(port_info['removed']):
            self._treat_devices_removed(port_info['removed'])

    def daemon_loop(self):
        ports = set()
        # Get all exsited ports as configured
        all_ports_info = self._update_ports(ports)
        if all_ports_info is not None:
            ports = all_ports_info['current']

        while True:
            try:
                start_time = time.time()
                port_info = self._update_ports(ports)

                if port_info:
                    LOG.info("Devices change, info: %s", port_info)
                    self._process_network_ports(port_info)
                    ports = port_info['current']
            except Exception as e:
                LOG.exception("Error in neutron agent loop: %s", e)

            # sleep till end of polling interval
            elapsed = (time.time() - start_time)
            if (elapsed < self._polling_interval):
                sleep_time = self._polling_interval - elapsed
                LOG.debug("Sleep %s", sleep_time)
                time.sleep(sleep_time)
            else:
                LOG.debug("Looping iteration exceeded interval")

    @restart_wrapper
    def _handle_restart(self):
        zvm_uptime = None
        while True:
            LOG.info("Try to reinitialize network ... ")
            try:
                tmp_new_time = self._requesthandler.call('host_get_info').get(
                                        'ipl_time')
                if zvm_uptime != tmp_new_time:
                    self._port_map = self._utils.get_port_map()
                    zvm_uptime = tmp_new_time
                yield
            except Exception:
                LOG.error("Failed to handle restart,"
                          "try again in 5 seconds")
                time.sleep(5)


def main():
    eventlet.monkey_patch()
    CONF(project='neutron')
    common_config.init(sys.argv[1:])
    common_config.setup_logging()

    agent = zvmNeutronAgent()

    # Start to query ZVMSDK DB
    LOG.info("z/VM agent initialized, now running... ")
    agent.daemon_loop()
    sys.exit(0)



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the z/VM network.
"""

import mock
from oslo_config import cfg

from neutron.plugins.zvm.agent import zvm_network
from neutron.plugins.zvm.common import utils as zvmutils
from neutron.tests import base


SDK_URL = 'https://10.10.10.1:8080'
FLAT_NETWORKS = ['flat_net1', '9dotvsw']
VLAN_NETWORKS = ['vlan_net1:100:500', '10dotvsw:10:100']
NETWORK_MAPS = {'vlan_net1': [(100, 500)], 'flat_net1': [],
                '9dotvsw': [], '10dotvsw': [(10, 100)]}


class TestZVMNetwork(base.BaseTestCase):

    _FAKE_NETWORK_VLAN_RANGES = "fakevsw1:1:4094,fakevsw2,fakevsw3:2:2999"

    @mock.patch.object(zvmutils.zVMConnectorRequestHandler, 'call')
    def setUp(self, call):
        super(TestZVMNetwork, self).setUp()
        cfg.CONF.set_override('flat_networks', FLAT_NETWORKS,
                              group='ml2_type_flat')
        cfg.CONF.set_override('network_vlan_ranges', VLAN_NETWORKS,
                              group='ml2_type_vlan')
        cfg.CONF.set_override('cloud_connector_url', SDK_URL,
                              group='AGENT')
        call.return_value = []
        self._zvm_network = zvm_network.zvmNetwork()

    def test_init_driver(self):
        self.assertIsInstance(self._zvm_network._requesthandler,
                              zvmutils.zVMConnectorRequestHandler)

    def test_get_network_maps(self):
        maps = self._zvm_network.get_network_maps()
        self.assertEqual(maps, NETWORK_MAPS)



# Copyright 2014 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import utils as plugin_utils
from oslo_config import cfg
from oslo_log import log as logging

from neutron.plugins.zvm.common import utils

LOG = logging.getLogger(__name__)

vswitch_opts = [
    cfg.StrOpt('rdev_list',
               help='RDev list for vswitch uplink port')]

CONF = cfg.CONF
CONF.import_opt('flat_networks', "neutron.plugins.ml2.drivers.type_flat",
                'ml2_type_flat')
CONF.import_opt('network_vlan_ranges', "neutron.plugins.ml2.drivers.type_vlan",
                'ml2_type_vlan')


class zvmVswitch(object):
    def __init__(self, name, vlan):
        self._requesthandler = utils.zVMConnectorRequestHandler()
        # check vlan set
        if not len(vlan):
            vlan = 'UNAWARE'
        else:
            vlan = str(vlan[0][0])
        self._requesthandler.call('vswitch_create', name,
                          rdev=getattr(CONF.get(name), "rdev_list"),
                          vid=vlan, network_type='ETHERNET')


class zvmNetwork(object):
    def __init__(self):
        self._requesthandler = utils.zVMConnectorRequestHandler()
        self._vsws = []
        self._maps = {}
        self._creat_networks()

    def _creat_networks(self):
        admin_vsw = self._requesthandler.call('vswitch_get_list')
        self._maps = plugin_utils.parse_network_vlan_ranges(
                            CONF.ml2_type_vlan.network_vlan_ranges +
                            CONF.ml2_type_flat.flat_networks)
        self._vsws = []
        for vsw in self._maps.keys():
            CONF.register_opts(vswitch_opts, vsw)
            if vsw.upper() in admin_vsw:
                LOG.info('Vswitch %s is pre-created by admin or system, '
                    'neutron-zvm-agent will not handle it' % vsw)
            else:
                self._vsws.append(zvmVswitch(vsw, self._maps[vsw]))

    def get_network_maps(self):
        return self._maps



# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron.plugins.ml2.drivers.zvm import mech_zvm
from neutron.tests.unit.plugins.ml2 import _test_mech_agent as base


class ZvmMechanismBaseTestCase(base.AgentMechanismBaseTestCase):
    VIF_TYPE = mech_zvm.VIF_TYPE_ZVM
    CAP_PORT_FILTER = False
    AGENT_TYPE = mech_zvm.AGENT_TYPE_ZVM

    GOOD_MAPPINGS = {'fake_physical_network': 'fake_vswitch'}
    GOOD_CONFIGS = {'vswitch_mappings': GOOD_MAPPINGS}

    BAD_MAPPINGS = {'wrong_physical_network': 'wrong_vswitch'}
    BAD_CONFIGS = {'vswitch_mappings': BAD_MAPPINGS}

    AGENTS = [{'alive': True,
               'configurations': GOOD_CONFIGS,
               'host': 'host'}]
    AGENTS_DEAD = [{'alive': False,
                    'configurations': GOOD_CONFIGS,
                    'host': 'dead_host'}]
    AGENTS_BAD = [{'alive': False,
                   'configurations': GOOD_CONFIGS,
                   'host': 'bad_host_1'},
                  {'alive': True,
                   'configurations': BAD_CONFIGS,
                   'host': 'bad_host_2'}]

    def setUp(self):
        super(ZvmMechanismBaseTestCase, self).setUp()
        self.driver = mech_zvm.ZvmMechanismDriver()
        self.driver.initialize()


class ZvmMechanismGenericTestCase(ZvmMechanismBaseTestCase,
                                  base.AgentMechanismGenericTestCase):
    pass


class ZvmMechanismLocalTestCase(ZvmMechanismBaseTestCase,
                                base.AgentMechanismLocalTestCase):
    pass


class ZvmvMechanismFlatTestCase(ZvmMechanismBaseTestCase,
                                base.AgentMechanismFlatTestCase):
    pass


class ZvmvMechanismVlanTestCase(ZvmMechanismBaseTestCase,
                                base.AgentMechanismVlanTestCase):
    pass



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging
import six.moves.urllib.parse as urlparse

from neutron.plugins.zvm.common import config as cfg
from neutron.plugins.zvm.common import exception
from zvmconnector import connector

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class zVMConnectorRequestHandler(object):

    def __init__(self):
        _url = urlparse.urlparse(CONF.AGENT.cloud_connector_url)
        _ca_file = CONF.AGENT.zvm_cloud_connector_ca_file
        _token_file = CONF.AGENT.zvm_cloud_connector_token_file
        kwargs = {}
        if _url.scheme == 'https':
            kwargs['ssl_enabled'] = True
        else:
            kwargs['ssl_enabled'] = False

        if _token_file is not None:
            kwargs['token_path'] = _token_file

        if ((kwargs['ssl_enabled']) and
            (_ca_file is not None)):
            kwargs['verify'] = _ca_file
        else:
            kwargs['verify'] = False

        self._conn = connector.ZVMConnector(_url.hostname, _url.port, **kwargs)

    def call(self, func_name, *args, **kwargs):
        results = self._conn.send_request(func_name, *args, **kwargs)
        if results['overallRC'] == 0:
            return results['output']
        else:
            msg = ("SDK request %(api)s failed with parameters: %(args)s "
                   "%(kwargs)s . Results: %(results)s" %
                   {'api': func_name, 'args': str(args), 'kwargs': str(kwargs),
                    'results': str(results)})
            LOG.debug(msg)
            raise exception.ZVMSDKRequestFailed(msg=msg, results=results)


class zvmUtils(object):

    def __init__(self):
        self._requesthandler = zVMConnectorRequestHandler()

    def get_port_map(self):
        ports_info = self._requesthandler.call('guests_get_nic_info')
        ports = {}
        for p in ports_info:
            if p['port'] is not None:
                userid = p['userid']
                vswitch = p['switch']
                port_id = p['port']
                ports[port_id] = {'userid': userid,
                                  'vswitch': vswitch}
        return ports



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from neutron_lib import exceptions as exception

from neutron._i18n import _


class zVMInvalidDataError(exception.NeutronException):
    message = _('Invalid data error: %(msg)s')


class ZVMSDKRequestFailed(exception.NeutronException):
    message = _('z/VM SDK request failed: %(msg)s')

    def __init__(self, **kwargs):
        self.results = {'rs': 0, 'overallRC': 1, 'modID': 0, 'rc': 0,
                        'output': '', 'errmsg': ''}
        if 'results' in kwargs:
            self.results = kwargs.pop('results')
        super(ZVMSDKRequestFailed, self).__init__(**kwargs)



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the z/VM utils.
"""
import mock
import os

from oslo_config import cfg

from neutron.plugins.zvm.common import exception
from neutron.plugins.zvm.common import utils
from neutron.tests import base
from zvmconnector import connector

HTTPS_SDK_URL = 'https://10.10.10.1:8080'
HTTP_SDK_URL = 'http://10.10.10.1:8080'
CA_FILE = '/tmp/ca.pem'
TOKEN_FILE = '/tmp/token.dat'


class TestZVMUtils_HTTPS_without_verify(base.BaseTestCase):
    def setUp(self):
        super(TestZVMUtils_HTTPS_without_verify, self).setUp()
        self.addCleanup(cfg.CONF.reset)
        cfg.CONF.set_override('cloud_connector_url', HTTPS_SDK_URL,
                              group='AGENT')
        self._utils = utils.zVMConnectorRequestHandler()

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call(self, request):
        request.return_value = {"overallRC": 0, 'output': "OK"}
        info = self._utils.call('API', "parm1", "parm2")
        request.assert_called_with('API', "parm1", "parm2")
        self.assertEqual("OK", info)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call_exception(self, request):
        request.return_value = {"overallRC": 1, 'output': ""}
        self.assertRaises(exception.ZVMSDKRequestFailed,
                          self._utils.call,
                          "API")


class TestZVMUtils_HTTPS_with_verify(base.BaseTestCase):
    def setUp(self):
        super(TestZVMUtils_HTTPS_with_verify, self).setUp()
        if not os.path.exists(CA_FILE):
            os.mknod(CA_FILE)
        cfg.CONF.set_override('cloud_connector_url', HTTPS_SDK_URL,
                              group='AGENT')
        cfg.CONF.set_override('zvm_cloud_connector_ca_file', CA_FILE,
                              group='AGENT')
        self._utils = utils.zVMConnectorRequestHandler()
        self.addCleanup(cfg.CONF.reset)
        self.addCleanup(os.remove, CA_FILE)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call(self, request):
        request.return_value = {"overallRC": 0, 'output': "OK"}
        info = self._utils.call('API', "parm1", "parm2")
        request.assert_called_with('API', "parm1", "parm2")
        self.assertEqual("OK", info)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call_exception(self, request):
        request.return_value = {"overallRC": 1, 'output': ""}
        self.assertRaises(exception.ZVMSDKRequestFailed,
                          self._utils.call,
                          "API")


class TestZVMUtils_HTTP_without_verify(base.BaseTestCase):
    def setUp(self):
        super(TestZVMUtils_HTTP_without_verify, self).setUp()
        self.addCleanup(cfg.CONF.reset)
        cfg.CONF.set_override('cloud_connector_url', HTTP_SDK_URL,
                              group='AGENT')
        self._utils = utils.zVMConnectorRequestHandler()

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call(self, request):
        request.return_value = {"overallRC": 0, 'output': "OK"}
        info = self._utils.call('API', "parm1", "parm2")
        request.assert_called_with('API', "parm1", "parm2")
        self.assertEqual("OK", info)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call_exception(self, request):
        request.return_value = {"overallRC": 1, 'output': ""}
        self.assertRaises(exception.ZVMSDKRequestFailed,
                          self._utils.call,
                          "API")


class TestZVMUtils_HTTP_with_verify(base.BaseTestCase):
    def setUp(self):
        super(TestZVMUtils_HTTP_with_verify, self).setUp()
        if not os.path.exists(CA_FILE):
            os.mknod(CA_FILE)
        cfg.CONF.set_override('cloud_connector_url', HTTP_SDK_URL,
                              group='AGENT')
        cfg.CONF.set_override('zvm_cloud_connector_ca_file', CA_FILE,
                              group='AGENT')
        self._utils = utils.zVMConnectorRequestHandler()
        self.addCleanup(cfg.CONF.reset)
        self.addCleanup(os.remove, CA_FILE)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call(self, request):
        request.return_value = {"overallRC": 0, 'output': "OK"}
        info = self._utils.call('API', "parm1", "parm2")
        request.assert_called_with('API', "parm1", "parm2")
        self.assertEqual("OK", info)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call_exception(self, request):
        request.return_value = {"overallRC": 1, 'output': ""}
        self.assertRaises(exception.ZVMSDKRequestFailed,
                          self._utils.call,
                          "API")


class TestZVMUtils_HTTP_token_with_verify(base.BaseTestCase):
    def setUp(self):
        super(TestZVMUtils_HTTP_token_with_verify, self).setUp()
        if not os.path.exists(CA_FILE):
            os.mknod(CA_FILE)
        if not os.path.exists(TOKEN_FILE):
            os.mknod(TOKEN_FILE)
        cfg.CONF.set_override('cloud_connector_url', HTTP_SDK_URL,
                              group='AGENT')
        cfg.CONF.set_override('zvm_cloud_connector_ca_file', CA_FILE,
                              group='AGENT')
        cfg.CONF.set_override('zvm_cloud_connector_token_file', TOKEN_FILE,
                              group='AGENT')
        self._utils = utils.zVMConnectorRequestHandler()
        self.addCleanup(cfg.CONF.reset)
        self.addCleanup(os.remove, CA_FILE)
        self.addCleanup(os.remove, TOKEN_FILE)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call(self, request):
        request.return_value = {"overallRC": 0, 'output': "OK"}
        info = self._utils.call('API', "parm1", "parm2")
        request.assert_called_with('API', "parm1", "parm2")
        self.assertEqual("OK", info)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call_exception(self, request):
        request.return_value = {"overallRC": 1, 'output': ""}
        self.assertRaises(exception.ZVMSDKRequestFailed,
                          self._utils.call,
                          "API")


class TestZVMUtils_HTTPS_token_without_verify(base.BaseTestCase):
    def setUp(self):
        super(TestZVMUtils_HTTPS_token_without_verify, self).setUp()
        if not os.path.exists(TOKEN_FILE):
            os.mknod(TOKEN_FILE)
        cfg.CONF.set_override('cloud_connector_url', HTTPS_SDK_URL,
                              group='AGENT')
        cfg.CONF.set_override('zvm_cloud_connector_token_file', TOKEN_FILE,
                              group='AGENT')
        self._utils = utils.zVMConnectorRequestHandler()
        self.addCleanup(cfg.CONF.reset)
        self.addCleanup(os.remove, TOKEN_FILE)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call(self, request):
        request.return_value = {"overallRC": 0, 'output': "OK"}
        info = self._utils.call('API', "parm1", "parm2")
        request.assert_called_with('API', "parm1", "parm2")
        self.assertEqual("OK", info)

    @mock.patch.object(connector.ZVMConnector, 'send_request')
    def test_call_exception(self, request):
        request.return_value = {"overallRC": 1, 'output': ""}
        self.assertRaises(exception.ZVMSDKRequestFailed,
                          self._utils.call,
                          "API")
